# -*- coding: utf-8 -*-

"""CLI utils."""

import json
import os
from collections import OrderedDict
from pathlib import Path
from typing import Dict, TextIO, Union

import click
import pandas as pd

from pykeen.constants import KG_EMBEDDING_MODEL_NAME

__all__ = [
    'get_config_dict',
    'summarize_results',
]


def get_config_dict(model_name: str) -> Dict:
    """Get configuration dictionary."""
    return OrderedDict([
        (KG_EMBEDDING_MODEL_NAME, model_name),
    ])


def summarize_results(directory: str, output: Union[str, Path, TextIO]) -> None:
    """Summarize contents of training and evaluation."""
    r = []

    for subdirectory_name in os.listdir(directory):
        subdirectory = os.path.join(directory, subdirectory_name)
        if not os.path.isdir(subdirectory):
            continue

        configuration_path = os.path.join(subdirectory, 'configuration.json')
        if not os.path.exists(configuration_path):
            click.echo("missing configuration")
            continue
        with open(configuration_path) as file:
            configuration = json.load(file)

        evaluation_path = os.path.join(subdirectory, 'evaluation_summary.json')
        if not os.path.exists(evaluation_path):
            click.echo("missing evaluation summary")
            continue
        with open(evaluation_path) as file:
            evaluation = json.load(file)

        r.append(dict(**configuration, **evaluation))

    df = pd.DataFrame(r)
    df.to_csv(output, sep='\t')



# -*- coding: utf-8 -*-

"""Utilities for getting and initializing KGE models."""

from typing import Dict

from torch.nn import Module

from pykeen.constants import KG_EMBEDDING_MODEL_NAME
from pykeen.kge_models import (
    ConvE, DistMult, ERMLP, RESCAL, StructuredEmbedding, TransD, TransE, TransH, TransR, UnstructuredModel,
)

__all__ = [
    'KGE_MODELS',
    'get_kge_model',
]

#: A mapping from KGE model names to KGE model classes
KGE_MODELS = {
    TransE.model_name: TransE,
    TransH.model_name: TransH,
    TransD.model_name: TransD,
    TransR.model_name: TransR,
    StructuredEmbedding.model_name: StructuredEmbedding,
    UnstructuredModel.model_name: UnstructuredModel,
    DistMult.model_name: DistMult,
    ERMLP.model_name: ERMLP,
    RESCAL.model_name: RESCAL,
    ConvE.model_name: ConvE,
}


def get_kge_model(config: Dict) -> Module:
    """Get an instance of a knowledge graph embedding model with the given configuration."""
    kge_model_name = config[KG_EMBEDDING_MODEL_NAME]
    kge_model_cls = KGE_MODELS.get(kge_model_name)

    if kge_model_cls is None:
        raise ValueError(f'Invalid KGE model name: {kge_model_name}')

    return kge_model_cls(config=config)



# -*- coding: utf-8 -*-

"""Implementation of the basic pipeline."""

import logging
from collections import OrderedDict
from typing import Dict, Iterable, Mapping, Optional, Tuple, Union

import numpy as np
import torch
from sklearn.model_selection import train_test_split
from torch.nn import Module

import pykeen.constants as pkc
from pykeen.hyper_parameter_optimizer.random_search_optimizer import RandomSearchHPO
from pykeen.kge_models import get_kge_model
from pykeen.utilities.evaluation_utils.metrics_computations import MetricResults, compute_metric_results
from pykeen.utilities.train_utils import train_kge_model
from pykeen.utilities.triples_creation_utils import create_mapped_triples, create_mappings

__all__ = [
    'Pipeline',
    'load_data',
]

log = logging.getLogger(__name__)


class Pipeline(object):
    """Encapsulates the KGE model training pipeline."""

    def __init__(self, config: Dict):
        self.config: Dict = config
        self.seed: int = config[pkc.SEED] if pkc.SEED in config else 2
        self.entity_label_to_id: Dict[int: str] = None
        self.relation_label_to_id: Dict[int: str] = None
        self.device_name = (
            'cuda:0'
            if torch.cuda.is_available() and self.config[pkc.PREFERRED_DEVICE] == pkc.GPU else
            pkc.CPU
        )
        self.device = torch.device(self.device_name)

    @staticmethod
    def _use_hpo(config):
        return config[pkc.EXECUTION_MODE] == pkc.HPO_MODE

    @property
    def is_evaluation_required(self) -> bool:
        return pkc.TEST_SET_PATH in self.config or pkc.TEST_SET_RATIO in self.config

    def run(self) -> Mapping:
        """Run this pipeline."""
        metric_results = None

        if self._use_hpo(self.config):  # Hyper-parameter optimization mode
            mapped_pos_train_triples, mapped_pos_test_triples = self._get_train_and_test_triples()

            (trained_model,
             loss_per_epoch,
             entity_label_to_embedding,
             relation_label_to_embedding,
             eval_summary,
             params) = RandomSearchHPO.run(
                mapped_train_triples=mapped_pos_train_triples,
                mapped_test_triples=mapped_pos_test_triples,
                entity_to_id=self.entity_label_to_id,
                rel_to_id=self.relation_label_to_id,
                config=self.config,
                device=self.device,
                seed=self.seed,
            )
        else:  # Training Mode
            if self.is_evaluation_required:
                mapped_pos_train_triples, mapped_pos_test_triples = self._get_train_and_test_triples()
            else:
                mapped_pos_train_triples = self._get_train_triples()

            all_entities = np.array(list(self.entity_label_to_id.values()))

            # Initialize KG embedding model
            self.config[pkc.NUM_ENTITIES] = len(self.entity_label_to_id)
            self.config[pkc.NUM_RELATIONS] = len(self.relation_label_to_id)
            self.config[pkc.PREFERRED_DEVICE] = pkc.CPU if self.device_name == pkc.CPU else pkc.GPU
            if self.seed is not None:
                torch.manual_seed(self.seed)

            kge_model: Module = get_kge_model(config=self.config)

            batch_size = self.config[pkc.BATCH_SIZE]
            num_epochs = self.config[pkc.NUM_EPOCHS]
            learning_rate = self.config[pkc.LEARNING_RATE]

            log.info("-------------Train KG Embeddings-------------")
            trained_model, loss_per_epoch = train_kge_model(
                kge_model=kge_model,
                all_entities=all_entities,
                learning_rate=learning_rate,
                num_epochs=num_epochs,
                batch_size=batch_size,
                pos_triples=mapped_pos_train_triples,
                device=self.device,
                seed=self.seed,
            )

            params = self.config

            if self.is_evaluation_required:
                log.info("-------------Start Evaluation-------------")
                metric_results = compute_metric_results(
                    all_entities=all_entities,
                    kg_embedding_model=kge_model,
                    mapped_train_triples=mapped_pos_train_triples,
                    mapped_test_triples=mapped_pos_test_triples,
                    device=self.device,
                    filter_neg_triples=self.config[pkc.FILTER_NEG_TRIPLES],
                )

        # Prepare Output
        entity_id_to_label = {
            value: key
            for key, value in self.entity_label_to_id.items()
        }
        relation_id_to_label = {
            value: key for
            key, value in self.relation_label_to_id.items()
        }
        entity_label_to_embedding = {
            entity_id_to_label[entity_id]: embedding.detach().cpu().numpy()
            for entity_id, embedding in enumerate(trained_model.entity_embeddings.weight)
        }

        if self.config[pkc.KG_EMBEDDING_MODEL_NAME] in (pkc.SE_NAME, pkc.UM_NAME):
            relation_label_to_embedding = None
        else:
            relation_label_to_embedding = {
                relation_id_to_label[relation_id]: embedding.detach().cpu().numpy()
                for relation_id, embedding in enumerate(trained_model.relation_embeddings.weight)
            }

        return _make_results(
            trained_model=trained_model,
            loss_per_epoch=loss_per_epoch,
            entity_to_embedding=entity_label_to_embedding,
            relation_to_embedding=relation_label_to_embedding,
            metric_results=metric_results,
            entity_to_id=self.entity_label_to_id,
            rel_to_id=self.relation_label_to_id,
            params=params,
        )

    def _get_train_and_test_triples(self) -> Tuple[np.ndarray, np.ndarray]:
        train_pos = load_data(self.config[pkc.TRAINING_SET_PATH])

        if pkc.TEST_SET_PATH in self.config:
            test_pos = load_data(self.config[pkc.TEST_SET_PATH])
        else:
            train_pos, test_pos = train_test_split(
                train_pos,
                test_size=self.config.get(pkc.TEST_SET_RATIO, 0.1),
                random_state=self.seed,
            )

        return self._handle_train_and_test(train_pos, test_pos)

    def _handle_train_and_test(self, train_pos, test_pos) -> Tuple[np.ndarray, np.ndarray]:
        """"""
        all_triples: np.ndarray = np.concatenate([train_pos, test_pos], axis=0)
        self.entity_label_to_id, self.relation_label_to_id = create_mappings(triples=all_triples)

        mapped_pos_train_triples, _, _ = create_mapped_triples(
            triples=train_pos,
            entity_label_to_id=self.entity_label_to_id,
            relation_label_to_id=self.relation_label_to_id,
        )

        mapped_pos_test_triples, _, _ = create_mapped_triples(
            triples=test_pos,
            entity_label_to_id=self.entity_label_to_id,
            relation_label_to_id=self.relation_label_to_id,
        )

        return mapped_pos_train_triples, mapped_pos_test_triples

    def _get_train_triples(self):
        train_pos = load_data(self.config[pkc.TRAINING_SET_PATH])

        self.entity_label_to_id, self.relation_label_to_id = create_mappings(triples=train_pos)

        mapped_pos_train_triples, _, _ = create_mapped_triples(
            triples=train_pos,
            entity_label_to_id=self.entity_label_to_id,
            relation_label_to_id=self.relation_label_to_id,
        )

        return mapped_pos_train_triples


def load_data(path: Union[str, Iterable[str]]) -> np.ndarray:
    """Load data given the *path*."""
    if isinstance(path, str):
        return _load_data_helper(path)

    return np.concatenate([
        _load_data_helper(p)
        for p in path
    ])


def _load_data_helper(path: str) -> np.ndarray:
    for prefix, handler in pkc.IMPORTERS.items():
        if path.startswith(f'{prefix}:'):
            return handler(path[len(f'{prefix}:'):])

    if path.endswith('.tsv'):
        return np.reshape(np.loadtxt(
            fname=path,
            dtype=str,
            comments='@Comment@ Subject Predicate Object',
            delimiter='\t',
        ), newshape=(-1, 3))

    if path.endswith('.nt'):
        import rdflib
        g = rdflib.Graph()
        g.parse(path, format='nt')
        return np.array(
            [
                [str(s), str(p), str(o)]
                for s, p, o in g
            ],
            dtype=np.str,
        )

    raise ValueError('''The argument to _load_data must be one of the following:

    - A string path to a .tsv file containing 3 columns corresponding to subject, predicate, and object
    - A string path to a .nt RDF file serialized in N-Triples format
    - A string NDEx network UUID prefixed by "ndex:" like in ndex:f93f402c-86d4-11e7-a10d-0ac135e8bacf
    ''')


def _make_results(trained_model,
                  loss_per_epoch,
                  entity_to_embedding: Mapping[str, np.ndarray],
                  relation_to_embedding: Mapping[str, np.ndarray],
                  metric_results: Optional[MetricResults],
                  entity_to_id,
                  rel_to_id,
                  params) -> Dict:
    results = OrderedDict()
    results[pkc.TRAINED_MODEL] = trained_model
    results[pkc.LOSSES] = loss_per_epoch
    results[pkc.ENTITY_TO_EMBEDDING]: Mapping[str, np.ndarray] = entity_to_embedding
    results[pkc.RELATION_TO_EMBEDDING]: Mapping[str, np.ndarray] = relation_to_embedding
    if metric_results is not None:
        results[pkc.EVAL_SUMMARY] = {
            pkc.MEAN_RANK: metric_results.mean_rank,
            pkc.HITS_AT_K: metric_results.hits_at_k,
        }
    results[pkc.ENTITY_TO_ID] = entity_to_id
    results[pkc.RELATION_TO_ID] = rel_to_id
    results[pkc.FINAL_CONFIGURATION] = params
    return results



# -*- coding: utf-8 -*-

"""PyKEEN's command line interface helper."""

import click


def print_section_divider(width: int = 104) -> None:
    """Print the section divider."""
    click.clear()
    click.secho('-' * width)


def print_welcome_message(width: int = 49, name='PyKEEN') -> None:
    """Print the PyKEEN welcome message."""
    click.echo('#' * 49)
    click.secho("#\t\tWelcome to " + click.style(name, bold=True) + "\t\t#")
    click.echo('#' * 49)


def print_intro() -> None:
    """Print the PyKEEN introduction."""
    click.secho("This interface will assist you to configure your experiment.\n")
    click.secho(
        "PyKEEN can be run in two modes: \n"
        "1.) Training mode: PyKEEN trains a model based on a set of user-defined hyper-parameters.\n"
        "2.) Hyper-parameter optimization mode: "
        "Apply Random Search to determine the most appropriate set of hyper-parameter values"
    )


def print_existing_config_message():
    click.secho(
        "Here you are asked whether you have already an existing configuration that was created for a previous "
        "experiment.\nThe configuration is saved as a JSON file (.json)\n"
        "Example of a valid path: /Users/david/data/configuration.json\n"
    )


def print_training_set_message():
    click.secho(click.style("Current Step: Please provide the path to your training data.\n", fg='blue'))
    click.secho("The training file can be a TSV file (.tsv) containing three columns:", bold=True)
    click.echo("Column 1: Contains the subjects of the triples.")
    click.echo("Column 2: Contains the predicates of the triples.")
    click.echo("Column 3: Contains the objects of the triples.\n")
    click.echo("Example of a valid path: \"/Users/david/data/corpora/fb15k/fb_15k_train.tsv\"\n")
    click.secho("Alternatively, the training file can be a RDF file (.nt).\n", bold=True)
    click.echo("Example of a valid path: \"/Users/david/data/corpora/fb15k/fb_15k_train.net\"\n")
    click.secho("In addition, you can provide a string NDEx network UUID prefixed by \"ndex\" \n", bold=True)
    click.echo("Example of a valid NDEx network UUID: \"ndex:f93f402c-86d4-11e7-a10d-0ac135e8bacf\"\n")


def print_random_seed_message():
    click.secho(
        "Current Step: Please specify a random seed (positive integer).",
        fg='blue',
        bold=True,
    )
    click.secho("The random seed is used for negative sampling, and for the random train and test split if requested.")
    click.secho("")


def print_model_selection_message():
    click.secho("Current Step: Please choose one of the provided models.", fg='blue')
    click.echo(
        "Depending on which model you select, PyKEEN will assist you to configure the required hyper-parameters.\n")


def print_training_embedding_dimension_message():
    click.secho(
        "Current Step: Please specify the embedding dimension to use for learning the entities and relations.",
        fg='blue',
        bold=True,
    )
    click.secho("The embedding dimension must be a positive integer e.g. 20.\n")


def print_embedding_dimension_info_message():
    click.echo("For the selected model the embedding dimension of entities and relations are the same.\n")


def print_training_margin_loss_message():
    click.secho(
        "Current Step: Please specify the value of the margin loss used for the margin-ranking-loss function.",
        fg='blue',
        bold=True,
    )
    click.secho("The margin ranking loss is a float value. An example for the margin loss is a value of 1\n")


def print_scoring_fct_message():
    click.secho("Current Step: Please specify the norm used as scoring function.", fg='blue', bold=True)
    click.echo("The norm should be a positive integer value such as a value of 1\n")


def print_entity_normalization_message():
    click.secho("Current Step: Please specify the norm used to normalize the entities.", fg='blue', bold=True)
    click.echo("The norm should be a positive integer value such as a value of 2\n")


def print_learning_rate_message():
    click.secho("Current Step: Please specify the learning rate.", fg='blue', bold=True)
    click.secho("The learning rate should be a positive float value such as 0.01\n")


def print_batch_size_message():
    click.secho("Current Step: Please specify the batch size.", fg='blue', bold=True)
    click.secho("Typical batch sizes are 32,64 and 128\n")


def print_number_epochs_message():
    click.secho("Current Step: Please specify the number of epochs", fg='blue', bold=True)
    click.echo(
        "The number of epochs defines how often to iterate over the whole training set during the training the "
        "model.\n"
    )


def print_ask_for_evlauation_message():
    click.secho("Current Step: Please specify the number of epochs", fg='blue', bold=True)
    click.echo("Here you can specify whether you want to evaluate your model after training or not.\n")


def print_test_set_message():
    click.secho("Current Step: Please specify whether you provide a test set yourself, or whether the test set\n"
                "should be randomly extracted from the training set.\n", fg='blue', bold=True)


def print_test_ratio_message():
    click.secho("Current Step: Please specify the ratio of the training set that should be used as a test set",
                fg='blue', bold=True)
    click.echo('For example 0.5 means half of the training set is used as a test set.\n')


def print_filter_negative_triples_message():
    click.secho("Current Step: Please specify whether you want to filter negative triples out during evaluation.",
                fg='blue')
    click.echo('Filtered evaluation is more expressive, for further information we refer to \n'
               'Bordes et al. \"Translating embeddings for modeling multi-relational data.\"\n')


def print_output_directory_message():
    click.secho("Current Step: Please specify the path to your output directory.", fg='blue', bold=True)
    click.echo("Example of a valid path: /Users/david/output_direc\n")


def print_trans_h_soft_constraints_weight_message():
    click.secho("Current Step: Please specify the weight value for the soft constraints.", fg='blue', bold=True)
    click.secho('In TransH, soft constraints are introduced and incorporated into the loss function.\n'
                'For further information we refer to Wang, Zhen, et al. \"Knowledge Graph Embedding by Translating '
                'on Hyperplanes\n')


def print_entities_embedding_dimension_message():
    click.secho(
        "Current Step: Please specify the embedding dimension to use for "
        "learning the entities of the knowledge graph.\n",
        fg='blue'
    )


def print_relations_embedding_dimension_message():
    click.secho(
        "Current Step: Please specify the embedding dimension to use for "
        "learning the relations of the knowedge graph.\n",
        fg='blue'
    )


def print_conv_e_width_height_message():
    click.secho(
        "Current Step: Please specify the embedding width and height based "
        "on your specified embedding dimension.",
        fg='blue'
    )
    click.echo(
        "In ConvE, the input of the CNN is the embedding of the head and of "
        "the relation.\nThose are transformed into an \"image\" "
        "representation. The constraint is that height*width must equal\n"
        "to the embedding dimension. If the embedding dimension is for example "
        "100, then valid values for\nheight and width are e.g. 5 and 20 since "
        "5*20 = 100.\n"
    )


def print_conv_input_channels_message():
    click.secho("Current Step: Please specify the number of input channels for the convolution layer.\n", fg='blue',
                bold=True)


def print_conv_e_output_channels_message():
    click.secho("Current Step: Please specify the number of output channels for the convolution layer.\n", fg='blue',
                bold=True)


def print_conv_kernel_height_message():
    click.secho("Current Step: Please specify the height of the convolution kernel.", fg='blue', bold=True)
    click.echo("Important note: The kernel height must be smaller or equal to the input height, specified before.\n")


def print_conv_kernel_width_message():
    click.secho("Current Step: Please specify the width of the convolution kernel.", fg='blue', bold=True)
    click.echo("Important note: The kernel width must be smaller or equal to the input width, specified before.\n")


def print_input_dropout_message():
    click.secho("Current Step: Please specify the dropout rate for the input layer.", fg='blue', bold=True)
    click.secho("The dropout rate must be a value between 0 and 1\n")


def print_output_dropout_message():
    click.secho("Current Step: Please specify the dropout rate for the output layer.", fg='blue')
    click.secho("The dropout rate must be a value between 0 and 1\n")


def print_feature_map_dropout_message():
    click.secho(
        "Current Step: Please specify the dropout rate for the feature maps created by the convolution layer.",
        fg='blue', bold=True)
    click.secho("The dropout rate must be a value between 0 and 1\n")


def print_hpo_embedding_dimensions_message():
    click.secho("Current Step: Please specify a list of embedding dimensions for the entities and relations.",
                fg='blue', bold=True)
    click.secho("You can also provide just a single value, in this case the \',\' is not required.\n")


def print_hpo_margin_losses_message():
    click.secho("Current Step: Please provide a list of margin losses to use for the margin-ranking-loss function",
                fg='blue', bold=True)
    click.secho("The margin ranking losses need to be float values. Please separate your input by a \',\':\n"
                "0.5, 1, 2.4, etc.\n\n")


def print_hpo_scoring_fcts_message():
    click.secho("Current Step: Please provide a list of norms used as scoring function", fg='blue', bold=True)
    click.secho("The norms should be positive integer values. Please separate your input by a \',\':\n"
                "1, 2, 3, etc.\n\n")


def print_hpo_entity_normalization_norms_message():
    click.secho("Current Step: Please provide a list of norms used used for normalizing the entities.", fg='blue',
                bold=True)
    click.secho("The norms should be positive integer values. Please separate your input by a \',\':\n"
                "1, 2, 3, etc.\n\n")


def print_hpo_learning_rates_message():
    click.secho("Current Step: Please provide a list of learning rates", fg='blue', bold=True)
    click.secho("The learning rates need to be float values. Please separate your input by a \',\':\n"
                "0.1, 0.01, 0.001, etc.\n\n")


def print_hpo_batch_sizes_message():
    click.secho("Current Step: Please provide a list of batch sizes.", fg='blue', bold=True)
    click.secho("The batch sizes should be positive integer values. Please separate your input by a \',\':\n"
                "1, 2, 3, etc.\n\n")


def print_hpo_epochs_message():
    click.secho("Current Step: Please provide a list of epochs.", fg='blue', bold=True)
    click.secho("The epochs should be positive integer values. Please separate your input by a \',\':\n"
                "1, 2, 3, etc.\n\n")


def print_hpo_iterations_message():
    click.secho("Current Step: Please provide a list of epochs.", fg='blue', bold=True)
    click.secho("The epochs should be positive integer values. Please separate your input by a \',\':\n"
                "1, 2, 3, etc.\n\n")


def print_hpo_trans_h_soft_constraints_weights_message():
    click.secho("Current Step: Please provide a list of weight values for the soft constraints.", fg='blue', bold=True)
    click.secho('In TransH, soft constraints are introduced and incorporated into the loss function.\n'
                'For further information we refer to Wang, Zhen, et al. \"Knowledge Graph Embedding by Translating on Hyperplanes\n')


def print_hpo_conv_e_width_height_message():
    click.secho(
        click.style(
            "Current step: Please provide for each specified embedding dimension, the corrpespoinding embedding height and width.\n' \
            'Make sure that \'embedding dimension\' is equal to \'height * width\' ", fg='blue'))

    click.secho("In ConvE, the input of the CNN is the embedding of the head and of the relation.\n"
                "Those are transformed into an \"image\" representation. The constraint is that height*width must equal\n"
                "to the embedding dimension. If the embedding dimension is for example 100, then valid values for\n"
                "height and width are e.g. 5 and 20 since 5*20 = 100.\n")


def print_conv_e_hpo_kernel_height_message():
    click.secho(click.style("Current Step: Please specify the heights of the convolution kernel.", fg='blue'))
    click.echo("Important note: The kernel heights must be smaller or equal to the input heights specified before.\n")


def print_conv_e_hpo_kernel_width_message():
    click.secho(click.style("Current Step: Please specify the widths of the convolution kernel.", fg='blue'))
    click.echo("Important note: The kernel heights must be smaller or equal to the input heights specified before.\n")


def print_hpo_input_dropout_message():
    click.secho(click.style("Current Step: Please specify the dropout rates for the input layer.", fg='blue'))
    click.echo("The dropout rate must be a value between 0 and 1\n")


def print_hpo_output_dropout_message():
    click.secho(click.style("Current Step: Please specify the dropout rates for the output layer.", fg='blue'))
    click.echo("The dropout rate must be a value between 0 and 1\n")


def print_hpo_feature_maps_dropouts_message():
    click.secho(
        "Current Step: Please specify the dropout rates for the feature maps created by the convolution layer.",
        fg='blue')
    click.echo("The dropout rate must be a value between 0 and 1\n")


def print_optimizer_message():
    click.secho("Current Step: Please select the optimizer you want to use to train your model.\n", fg='blue')



# -*- coding: utf-8 -*-

"""Implementation the command line interface needed for TransE."""

from pykeen.cli.utils.cli_print_msg_helper import (
    print_batch_size_message, print_conv_e_hpo_kernel_height_message, print_conv_e_hpo_kernel_width_message,
    print_conv_e_output_channels_message, print_conv_e_width_height_message, print_conv_input_channels_message,
    print_conv_kernel_height_message, print_conv_kernel_width_message, print_embedding_dimension_info_message,
    print_feature_map_dropout_message, print_hpo_batch_sizes_message, print_hpo_conv_e_width_height_message,
    print_hpo_embedding_dimensions_message, print_hpo_epochs_message, print_hpo_feature_maps_dropouts_message,
    print_hpo_input_dropout_message, print_hpo_learning_rates_message, print_hpo_output_dropout_message,
    print_learning_rate_message, print_number_epochs_message, print_output_dropout_message, print_section_divider,
    print_training_embedding_dimension_message,
)
from pykeen.cli.utils.cli_query_helper import (
    query_height_and_width_for_conv_e, query_kernel_param, select_float_value, select_float_values,
    select_heights_and_widths, select_integer_value, select_kernel_sizes, select_positive_integer_values,
    select_zero_one_float_value, select_zero_one_range_float_values,
)
from pykeen.constants import (
    BATCH_SIZE, BATCH_SIZES_ERROR_MSG, BATCH_SIZES_PRINT_MSG, BATCH_SIZES_PROMPT_MSG,
    BATCH_SIZE_ERROR_MSG, BATCH_SIZE_PRINT_MSG, BATCH_SIZE_PROMPT_MSG, CONV_E_FEATURE_MAP_DROPOUT,
    CONV_E_FEATURE_MAP_DROPOUT_ERROR_MSG, CONV_E_FEATURE_MAP_DROPOUT_PRINT_MSG, CONV_E_HEIGHT,
    CONV_E_HPO_FEATURE_MAP_DROPOUT_ERROR_MSG, CONV_E_HPO_FEATURE_MAP_DROPOUT_PRINT_MSG,
    CONV_E_HPO_FEATURE_MAP_DROPOUT_PROMPT_MSG, CONV_E_HPO_INPUT_CHANNELS_ERROR_MSG, CONV_E_HPO_INPUT_CHANNELS_PRINT_MSG,
    CONV_E_HPO_INPUT_CHANNELS_PROMPT_MSG, CONV_E_HPO_INPUT_DROPOUTS_ERROR_MSG, CONV_E_HPO_INPUT_DROPOUTS_PRINT_MSG,
    CONV_E_HPO_INPUT_DROPOUTS_PROMPT_MSG, CONV_E_HPO_KERNEL_HEIGHTS_ERROR_MSG, CONV_E_HPO_KERNEL_HEIGHTS_PRINT_MSG,
    CONV_E_HPO_KERNEL_HEIGHTS_PROMPT_MSG, CONV_E_HPO_OUTPUT_DROPOUT_ERROR_MSG, CONV_E_HPO_OUTPUT_DROPOUT_PRINT_MSG,
    CONV_E_HPO_OUTPUT_DROPOUT_PROMPT_MSG, CONV_E_HPO_OUT_CHANNELS_ERROR_MSG, CONV_E_HPO_OUT_CHANNELS_PRINT_MSG,
    CONV_E_HPO_OUT_CHANNELS_PROMPT_MSG, CONV_E_INPUT_CHANNELS, CONV_E_INPUT_CHANNEL_ERROR_MSG,
    CONV_E_INPUT_CHANNEL_PRINT_MSG, CONV_E_INPUT_CHANNEL_PROMPT_MSG, CONV_E_INPUT_DROPOUT,
    CONV_E_INPUT_DROPOUT_ERROR_MSG, CONV_E_INPUT_DROPOUT_PRINT_MSG, CONV_E_INPUT_DROPOUT_PROMPT_MSG,
    CONV_E_KERNEL_HEIGHT, CONV_E_KERNEL_HEIGHT_ERROR_MSG, CONV_E_KERNEL_HEIGHT_PRINT_MSG,
    CONV_E_KERNEL_HEIGHT_PROMPT_MSG, CONV_E_KERNEL_WIDTH, CONV_E_KERNEL_WIDTH_ERROR_MSG, CONV_E_KERNEL_WIDTH_PRINT_MSG,
    CONV_E_KERNEL_WIDTH_PROMPT_MSG, CONV_E_OUTPUT_CHANNELS, CONV_E_OUTPUT_DROPOUT, CONV_E_OUTPUT_DROPOUT_ERROR_MSG,
    CONV_E_OUTPUT_DROPOUT_PRINT_MSG, CONV_E_OUTPUT_DROPOUT_PROMPT_MSG, CONV_E_OUT_CHANNEL_ERROR_MSG,
    CONV_E_OUT_CHANNEL_PRINT_MSG, CONV_E_OUT_CHANNEL_PROMPT_MSG, CONV_E_WIDTH, CONV_E__FEATURE_MAP_DROPOUT_PROMPT_MSG,
    EMBEDDING_DIM, EMBEDDING_DIMENSIONS_ERROR_MSG, EMBEDDING_DIMENSIONS_PRINT_MSG, EMBEDDING_DIMENSIONS_PROMPT_MSG,
    EMBEDDING_DIMENSION_ERROR_MSG, EMBEDDING_DIMENSION_PRINT_MSG, EMBEDDING_DIMENSION_PROMPT_MSG, EPOCHS_ERROR_MSG,
    EPOCHS_PRINT_MSG, EPOCHS_PROMPT_MSG, EPOCH_ERROR_MSG, EPOCH_PRINT_MSG, EPOCH_PROMPT_MSG, LEARNING_RATE,
    LEARNING_RATES_ERROR_MSG, LEARNING_RATES_PRINT_MSG, LEARNING_RATES_PROMPT_MSG, LEARNING_RATE_ERROR_MSG,
    LEARNING_RATE_PRINT_MSG, LEARNING_RATE_PROMPT_MSG, NUM_EPOCHS,
)
from pykeen.utilities.summarize import get_config_dict

__all__ = [
    'configure_conv_e_training_pipeline',
    'configure_conv_e_hpo_pipeline',
]


def configure_conv_e_training_pipeline(model_name: str):
    """Configure ConvE.

    :param str model_name: name of the model
    :return: configuration dictionary
    """
    config = get_config_dict(model_name)

    # Step 1: Query embedding dimension
    print_training_embedding_dimension_message()
    print_embedding_dimension_info_message()
    embedding_dimension = select_integer_value(
        print_msg=EMBEDDING_DIMENSION_PRINT_MSG,
        prompt_msg=EMBEDDING_DIMENSION_PROMPT_MSG,
        error_msg=EMBEDDING_DIMENSION_ERROR_MSG,
    )
    config[EMBEDDING_DIM] = embedding_dimension
    print_section_divider()

    # Step 2: Query height and width
    print_conv_e_width_height_message()
    height, width = query_height_and_width_for_conv_e(embedding_dimension)
    config[CONV_E_HEIGHT] = height
    config[CONV_E_WIDTH] = width
    print_section_divider()

    # Step 3: Query number of input channels
    print_conv_input_channels_message()
    num_input_channels = select_integer_value(
        CONV_E_INPUT_CHANNEL_PRINT_MSG,
        CONV_E_INPUT_CHANNEL_PROMPT_MSG,
        CONV_E_INPUT_CHANNEL_ERROR_MSG,
    )
    config[CONV_E_INPUT_CHANNELS] = num_input_channels
    print_section_divider()

    # Step 4: Query number of output channels
    print_conv_e_output_channels_message()
    num_output_channels = select_integer_value(
        CONV_E_OUT_CHANNEL_PRINT_MSG,
        CONV_E_OUT_CHANNEL_PROMPT_MSG,
        CONV_E_OUT_CHANNEL_ERROR_MSG,
    )
    config[CONV_E_OUTPUT_CHANNELS] = num_output_channels
    print_section_divider()

    # Step 4: Query kernel height
    print_conv_kernel_height_message()
    kernel_height = query_kernel_param(
        depending_param=height,
        print_msg=CONV_E_KERNEL_HEIGHT_PRINT_MSG,
        prompt_msg=CONV_E_KERNEL_HEIGHT_PROMPT_MSG,
        error_msg=CONV_E_KERNEL_HEIGHT_ERROR_MSG,
    )
    config[CONV_E_KERNEL_HEIGHT] = kernel_height
    print_section_divider()

    # Step 5: Query kernel width
    print_conv_kernel_width_message()
    kernel_width = query_kernel_param(
        depending_param=width,
        print_msg=CONV_E_KERNEL_WIDTH_PRINT_MSG,
        prompt_msg=CONV_E_KERNEL_WIDTH_PROMPT_MSG,
        error_msg=CONV_E_KERNEL_WIDTH_ERROR_MSG,
    )
    config[CONV_E_KERNEL_WIDTH] = kernel_width
    print_section_divider()

    # Step 6: Query dropout for input layer
    print_hpo_input_dropout_message()
    input_dropout = select_zero_one_float_value(
        print_msg=CONV_E_INPUT_DROPOUT_PRINT_MSG,
        prompt_msg=CONV_E_INPUT_DROPOUT_PROMPT_MSG,
        error_msg=CONV_E_INPUT_DROPOUT_ERROR_MSG,
    )
    config[CONV_E_INPUT_DROPOUT] = input_dropout
    print_section_divider()

    # Step 7: Query dropout for output layer
    print_output_dropout_message()
    output_dropout = select_zero_one_float_value(
        print_msg=CONV_E_OUTPUT_DROPOUT_PRINT_MSG,
        prompt_msg=CONV_E_OUTPUT_DROPOUT_PROMPT_MSG,
        error_msg=CONV_E_OUTPUT_DROPOUT_ERROR_MSG,
    )
    config[CONV_E_OUTPUT_DROPOUT] = output_dropout
    print_section_divider()

    # Step 8: Query feature map dropout for output layer
    print_feature_map_dropout_message()
    feature_map_dropout = select_zero_one_float_value(
        print_msg=CONV_E_FEATURE_MAP_DROPOUT_PRINT_MSG,
        prompt_msg=CONV_E__FEATURE_MAP_DROPOUT_PROMPT_MSG,
        error_msg=CONV_E_FEATURE_MAP_DROPOUT_ERROR_MSG,
    )
    config[CONV_E_FEATURE_MAP_DROPOUT] = feature_map_dropout
    print_section_divider()

    # Step 5: Query learning rate
    print_learning_rate_message()
    learning_rate = select_float_value(
        print_msg=LEARNING_RATE_PRINT_MSG,
        prompt_msg=LEARNING_RATE_PROMPT_MSG,
        error_msg=LEARNING_RATE_ERROR_MSG
    )
    config[LEARNING_RATE] = learning_rate
    print_section_divider()

    # Step 6: Query batch size
    print_batch_size_message()
    batch_size = select_integer_value(
        print_msg=BATCH_SIZE_PRINT_MSG,
        prompt_msg=BATCH_SIZE_PROMPT_MSG,
        error_msg=BATCH_SIZE_ERROR_MSG
    )
    config[BATCH_SIZE] = batch_size
    print_section_divider()

    # Step 7: Query number of epochs
    print_number_epochs_message()
    number_epochs = select_integer_value(
        print_msg=EPOCH_PRINT_MSG,
        prompt_msg=EPOCH_PROMPT_MSG,
        error_msg=EPOCH_ERROR_MSG
    )
    config[NUM_EPOCHS] = number_epochs
    print_section_divider()

    return config


def configure_conv_e_hpo_pipeline(model_name):
    """Configure ConvE.

    :param str model_name: name of the model
    :rtype: OrderedDict
    :return: configuration dictionary
    """
    config = get_config_dict(model_name)

    # Step 1: Query embedding dimension
    print_hpo_embedding_dimensions_message()
    embedding_dimensions = select_positive_integer_values(
        EMBEDDING_DIMENSIONS_PRINT_MSG,
        EMBEDDING_DIMENSIONS_PROMPT_MSG,
        EMBEDDING_DIMENSIONS_ERROR_MSG,
    )
    config[EMBEDDING_DIM] = embedding_dimensions
    print_section_divider()

    # Step 2: Query height and width
    print_hpo_conv_e_width_height_message()
    heights, widths = select_heights_and_widths(embedding_dimensions)
    config[CONV_E_HEIGHT] = heights
    config[CONV_E_WIDTH] = widths
    print_section_divider()

    # Step 3: Query number of input channels
    print_conv_input_channels_message()
    num_input_channels = select_positive_integer_values(
        CONV_E_HPO_INPUT_CHANNELS_PRINT_MSG,
        CONV_E_HPO_INPUT_CHANNELS_PROMPT_MSG,
        CONV_E_HPO_INPUT_CHANNELS_ERROR_MSG,
    )
    config[CONV_E_INPUT_CHANNELS] = num_input_channels
    print_section_divider()

    # Step 4: Query number of output channels
    print_conv_e_output_channels_message()
    num_output_channels = select_positive_integer_values(
        CONV_E_HPO_OUT_CHANNELS_PRINT_MSG,
        CONV_E_HPO_OUT_CHANNELS_PROMPT_MSG,
        CONV_E_HPO_OUT_CHANNELS_ERROR_MSG,
    )
    config[CONV_E_OUTPUT_CHANNELS] = num_output_channels
    print_section_divider()

    # Step 4: Query kernel height
    print_conv_e_hpo_kernel_height_message()
    kernel_heights = select_kernel_sizes(
        depending_params=heights,
        print_msg=CONV_E_HPO_KERNEL_HEIGHTS_PRINT_MSG,
        prompt_msg=CONV_E_HPO_KERNEL_HEIGHTS_PROMPT_MSG,
        error_msg=CONV_E_HPO_KERNEL_HEIGHTS_ERROR_MSG,
    )
    config[CONV_E_KERNEL_HEIGHT] = kernel_heights
    print_section_divider()

    # Step 5: Query kernel width
    print_conv_e_hpo_kernel_width_message()
    kernel_widths = select_kernel_sizes(
        depending_params=widths,
        print_msg=CONV_E_KERNEL_WIDTH_PRINT_MSG,
        prompt_msg=CONV_E_KERNEL_WIDTH_PROMPT_MSG,
        error_msg=CONV_E_KERNEL_WIDTH_ERROR_MSG,
    )
    config[CONV_E_KERNEL_WIDTH] = kernel_widths
    print_section_divider()

    # Step 6: Query dropout for input layer
    print_hpo_input_dropout_message()
    input_dropout = select_float_values(
        CONV_E_HPO_INPUT_DROPOUTS_PRINT_MSG,
        CONV_E_HPO_INPUT_DROPOUTS_PROMPT_MSG,
        CONV_E_HPO_INPUT_DROPOUTS_ERROR_MSG,
    )
    config[CONV_E_INPUT_DROPOUT] = input_dropout
    print_section_divider()

    # Step 7: Query dropout for output layer
    print_hpo_output_dropout_message()
    output_dropouts = select_zero_one_range_float_values(
        print_msg=CONV_E_HPO_OUTPUT_DROPOUT_PRINT_MSG,
        prompt_msg=CONV_E_HPO_OUTPUT_DROPOUT_PROMPT_MSG,
        error_msg=CONV_E_HPO_OUTPUT_DROPOUT_ERROR_MSG,
    )
    config[CONV_E_OUTPUT_DROPOUT] = output_dropouts
    print_section_divider()

    # Step 8: Query feature map dropout for output layer
    print_hpo_feature_maps_dropouts_message()
    feature_map_dropouts = select_zero_one_range_float_values(
        print_msg=CONV_E_HPO_FEATURE_MAP_DROPOUT_PRINT_MSG,
        prompt_msg=CONV_E_HPO_FEATURE_MAP_DROPOUT_PROMPT_MSG,
        error_msg=CONV_E_HPO_FEATURE_MAP_DROPOUT_ERROR_MSG,
    )
    config[CONV_E_FEATURE_MAP_DROPOUT] = feature_map_dropouts
    print_section_divider()

    # Step 9: Query learning rate
    print_hpo_learning_rates_message()
    learning_rates = select_float_values(
        print_msg=LEARNING_RATES_PRINT_MSG,
        prompt_msg=LEARNING_RATES_PROMPT_MSG,
        error_msg=LEARNING_RATES_ERROR_MSG,
    )
    config[LEARNING_RATE] = learning_rates
    print_section_divider()

    # Step 10: Query batch size
    print_hpo_batch_sizes_message()
    batch_size = select_positive_integer_values(
        print_msg=BATCH_SIZES_PRINT_MSG,
        prompt_msg=BATCH_SIZES_PROMPT_MSG,
        error_msg=BATCH_SIZES_ERROR_MSG,
    )
    config[BATCH_SIZE] = batch_size
    print_section_divider()

    # Step 11: Query number of epochs
    print_hpo_epochs_message()
    number_epochs = select_positive_integer_values(
        print_msg=EPOCHS_PRINT_MSG,
        prompt_msg=EPOCHS_PROMPT_MSG,
        error_msg=EPOCHS_ERROR_MSG,
    )
    config[NUM_EPOCHS] = number_epochs
    print_section_divider()

    return config



# -*- coding: utf-8 -*-

"""Handler for NDEx."""

import logging

import numpy as np

__all__ = ['handle_ndex']

log = logging.getLogger(__name__)


def handle_ndex(network_uuid: str) -> np.ndarray:
    """Load an NDEx network.

    Example network UUID: f93f402c-86d4-11e7-a10d-0ac135e8bacf
    """
    import ndex2
    ndex_client = ndex2.Ndex2()

    log.info(f'downloading {network_uuid} from ndex')
    res = ndex_client.get_network_as_cx_stream(network_uuid)
    res_json = res.json()

    triples = []

    for entry in res_json:
        for aspect, data in entry.items():
            if aspect == 'edges':
                for edge in data:
                    triples.append([
                        str(edge['s']),
                        edge.get('i', 'interacts'),
                        str(edge['t']),
                    ])

    return np.array(triples)



# -*- coding: utf-8 -*-

"""Implementation of RESCAL."""

from typing import Dict

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import RESCAL_NAME, SCORING_FUNCTION_NORM
from pykeen.kge_models.base import BaseModule, slice_triples

__all__ = ['RESCAL']


class RESCAL(BaseModule):
    """An implementation of RESCAL [nickel2011]_.

    This model represents relations as matrices and models interactions between latent features.

    .. [nickel2011] Nickel, M., *et al.* (2011) `A Three-Way Model for Collective Learning on Multi-Relational Data
                    <http://www.cip.ifi.lmu.de/~nickel/data/slides-icml2011.pdf>`_. ICML. Vol. 11.

    .. seealso::

       - Alternative implementation in OpenKE: https://github.com/thunlp/OpenKE/blob/master/models/RESCAL.py
    """

    model_name = RESCAL_NAME
    margin_ranking_loss_size_average: bool = True
    hyper_params = BaseModule.hyper_params + [SCORING_FUNCTION_NORM]

    def __init__(self, config: Dict) -> None:
        super().__init__(config)

        # Embeddings
        self.relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim * self.embedding_dim)

        self.scoring_fct_norm = config[SCORING_FUNCTION_NORM]

    def predict(self, triples):
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)
        scores = self._score_triples(triples)
        return scores.detach().cpu().numpy()

    def forward(self, positives, negatives):
        positive_scores = self._score_triples(positives)
        negative_scores = self._score_triples(negatives)
        loss = self._compute_loss(positive_scores, negative_scores)
        return loss

    def _compute_loss(self, positive_scores, negative_scores):
        # TODO: Check
        y = np.repeat([-1], repeats=positive_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        positive_scores = torch.tensor(positive_scores, dtype=torch.float, device=self.device)
        negative_scores = torch.tensor(negative_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(positive_scores, negative_scores, y)
        return loss

    def _score_triples(self, triples):
        head_embeddings, relation_embeddings, tail_embeddings = self._get_triple_embeddings(triples)
        scores = self._compute_scores(head_embeddings, relation_embeddings, tail_embeddings)
        return scores

    def _compute_scores(self, h_embs, r_embs, t_embs):
        # Compute score and transform result to 1D tensor
        m = r_embs.view(-1, self.embedding_dim, self.embedding_dim)
        h_embs = h_embs.unsqueeze(-1).permute([0, 2, 1])
        h_m_embs = torch.matmul(h_embs, m)
        t_embs = t_embs.unsqueeze(-1)
        scores = -torch.matmul(h_m_embs, t_embs).view(-1)

        # scores = torch.bmm(torch.transpose(h_emb, 1, 2), M)  # h^T M
        # scores = torch.bmm(scores, t_emb)  # (h^T M) h
        # scores = score.view(-1, 1)

        return scores

    def _get_triple_embeddings(self, triples):
        heads, relations, tails = slice_triples(triples)
        return (
            self._get_entity_embeddings(heads),
            self._get_relation_embeddings(relations),
            self._get_entity_embeddings(tails)
        )

    def _get_relation_embeddings(self, relations):
        return self.relation_embeddings(relations).view(-1, self.embedding_dim)



# -*- coding: utf-8 -*-

"""Constants defined for PyKEEN."""

from typing import Callable, Dict

import numpy as np
from pkg_resources import iter_entry_points

VERSION = '0.0.20'

#: Functions for specifying exotic resources with a given prefix
IMPORTERS: Dict[str, Callable[[str], np.ndarray]] = {
    entry_point.name: entry_point.load()
    for entry_point in iter_entry_points(group='pykeen.data.importer')
}


def get_version() -> str:
    """Get the version."""
    return VERSION


PYKEEN = 'PyKEEN'

# KG embedding model
KG_EMBEDDING_MODEL_NAME = 'kg_embedding_model_name'
EXECUTION_MODE = 'execution_mode'

# Model names
SE_NAME = 'SE'
UM_NAME = 'UM'
TRANS_E_NAME = 'TransE'
TRANS_H_NAME = 'TransH'
TRANS_D_NAME = 'TransD'
TRANS_R_NAME = 'TransR'
DISTMULT_NAME = 'DistMult'
ERMLP_NAME = 'ERMLP'
CONV_E_NAME = 'ConvE'
RESCAL_NAME = 'RESCAL'

# Evaluator
EVALUATOR = 'evaluator'
MEAN_RANK_EVALUATOR = 'MeanRankEvaluator'
RANDOM_SEARCH_OPTIMIZER = 'random_search_optimizer'
EVAL_METRICS = 'eval_metrics'
MEAN_RANK = 'mean_rank'
HITS_AT_K = 'hits@k'
FILTER_NEG_TRIPLES = 'filter_negative_triples'

# Output paths
ENTITY_TO_EMBEDDINGS = 'entity_to_embeddings'
EVAL_RESULTS = 'eval_results'
ENTITY_TO_ID = 'entity_to_id'
RELATION_TO_ID = 'relation_to_id'

# Device related
PREFERRED_DEVICE = 'preferred_device'
CPU = 'cpu'
GPU = 'gpu'

# ML params
BATCH_SIZE = 'batch_size'
VOCAB_SIZE = 'vocab_size'
EMBEDDING_DIM = 'embedding_dim'
RELATION_EMBEDDING_DIM = 'relation_embedding_dim'
MARGIN_LOSS = 'margin_loss'
NUM_ENTITIES = 'num_entities'
NUM_RELATIONS = 'num_relations'
NUM_EPOCHS = 'num_epochs'
NUM_OF_HPO_ITERS = 'maximum_number_of_hpo_iters'
LEARNING_RATE = 'learning_rate'
TRAINING_MODE = 'Training_mode'
HPO_MODE = 'HPO_mode'
HYPER_PARAMTER_OPTIMIZATION_PARAMS = 'hyper_optimization_params'
TRAINING_SET_PATH = 'training_set_path'
TEST_SET_PATH = 'test_set_path'
TEST_SET_RATIO = 'test_set_ratio'
NORM_FOR_NORMALIZATION_OF_ENTITIES = 'normalization_of_entities'
SCORING_FUNCTION_NORM = 'scoring_function'
# TransH related
WEIGHT_SOFT_CONSTRAINT_TRANS_H = 'weighting_soft_constraint'
# ConvE related
CONV_E_INPUT_DROPOUT = 'conv_e_input_dropout'
CONV_E_OUTPUT_DROPOUT = 'conv_e_output_dropout'
CONV_E_FEATURE_MAP_DROPOUT = 'conv_e_feature_map_dropout'
CONV_E_HEIGHT = 'ConvE_height'
CONV_E_WIDTH = 'ConvE_width'
CONV_E_INPUT_CHANNELS = 'ConvE_input_channels'
CONV_E_OUTPUT_CHANNELS = 'ConvE_output_channels'
CONV_E_KERNEL_HEIGHT = 'ConvE_kernel_height'
CONV_E_KERNEL_WIDTH = 'ConvE_kernel_width'

# OPTIMIZER
OPTMIZER_NAME = 'optimizer'
SGD_OPTIMIZER_NAME = 'SGD'
ADAGRAD_OPTIMIZER_NAME = 'Adagrad'
ADAM_OPTIMIZER_NAME = 'Adam'

# Further Constants
SEED = 'random_seed'
OUTPUT_DIREC = 'output_direc'

# Pipeline outcome parameters
TRAINED_MODEL = 'trained_model'
LOSSES = 'losses'
ENTITY_TO_EMBEDDING = 'entity_to_embedding'
RELATION_TO_EMBEDDING = 'relation_to_embedding'
CONFIG = 'configuration'
FINAL_CONFIGURATION = 'final_configuration'
EVAL_SUMMARY = 'eval_summary'
# -----------------Command line interface messages-----------------

TRAINING_FILE_PROMPT_MSG = '> Please provide here the path to your training file: '
TRAINING_FILE_ERROR_MSG = 'An error occurred, either the path is not correct or the training file doesn\'t exist.\n' \
                          'Please try again.'

TEST_FILE_PROMPT_MSG = '> Please provide here the path to your test file: '
TEST_FILE_ERROR_MSG = 'An error occurred, either the path is not correct or the test file doesn\'t exist.\n' \
                      'Please try again.'

EMBEDDING_DIMENSION_PRINT_MSG = 'Please type the range of preferred embedding dimensions for entities comma separated (e.g. 50,100,200):'
EMBEDDING_DIMENSION_PROMPT_MSG = '> Please select the embedding dimensions:'
EMBEDDING_DIMENSION_ERROR_MSG = 'An error occurred, please positive integer as embedding dimension.'

ENTITIES_EMBEDDING_DIMENSION_PRINT_MSG = 'Please provide the embedding dimension for entities (e.g. 40). '
ENTITIES_EMBEDDING_DIMENSION_PROMPT_MSG = '> Entity embedding dimension: '
ENTITIES_EMBEDDING_DIMENSION_ERROR_MSG = 'An error occurred, please proive and positive integer as embedding dimension (e.g. 50).'

ENTITIES_EMBEDDING_DIMENSIONS_PRINT_MSG = 'Please provide (comma separated) the embedding dimension(s) for entities (e.g. 50,100). '
ENTITIES_EMBEDDING_DIMENSIONS_PROMPT_MSG = '> Entity embedding dimensions: '
ENTITIES_EMBEDDING_DIMENSIONS_ERROR_MSG = 'An error occurred, the embedding dimensions must be positive integers and separated by a comma.\n' \
                                          'Please try again.'

BATCH_SIZES_PRINT_MSG = 'Please type (comma separated) the batch size(s) (e.g. 32, 64, 128):'
BATCH_SIZES_PROMPT_MSG = '> Batch size(s):'
BATCH_SIZES_ERROR_MSG = 'An error occurred, the batch sizes must be positive integers and separated by a comma.\n' \
                        'Please try again.'

EPOCHS_PRINT_MSG = 'Please type (comma separated) the number of epochs (e.g. 50,100,500).'
EPOCHS_PROMPT_MSG = '> Epochs:'
EPOCHS_ERROR_MSG = 'An error occurred, the number of epochs must be positive integers and separated by a comma.\n' \
                   'Please try again.'

LEARNING_RATES_PRINT_MSG = 'Please type (comma separated) the learning rate(s) (e.g. 0.1, 0.01, 0.0001).'
LEARNING_RATES_PROMPT_MSG = '> Learning rate(s):'
LEARNING_RATES_ERROR_MSG = 'An error occurred, the learning rates must be float values and separated by a comma.\n' \
                           'Please try again.'

MARGIN_LOSSES_PRINT_MSG = 'Please type (comma separated) the margin losses(s) comma separated  (e.g. 1,2,10), and press enter-'
MARGIN_LOSSES_PROMPT_MSG = '> Margin losse(s):'
MARGIN_LOSSES_ERROR_MSG = 'An error occurred, the margin losses must be positive float values and separated by a comma (e.g. 0.5,3)'

HPO_ITERS_PRINT_MSG = 'Please type (comma separated) the number of iterations of hyper-parameter search (e.g. 5)'
HPO_ITERS_PROMPT_MSG = '> Number of iterations: '
HPO_ITERS_ERROR_MSG = 'An error occurred, please type in a positive integer for the maximum number of iterations.'

EMBEDDING_DIMENSION_PRINT_MSG = 'Please provide the embedding dimension of entities and relations, and press enter.'
EMBEDDING_DIMENSION_PROMPT_MSG = '> Embedding dimension: '
EMBEDDING_DIMENSION_ERROR_MSG = 'An error occurred, please provide a positive integer as embedding dimension (e.g. 20).'

EMBEDDING_DIMENSIONS_PRINT_MSG = 'Please provide (comma separated) the embedding dimension(s), and press enter.'
EMBEDDING_DIMENSIONS_PROMPT_MSG = '> Embedding dimensions: '
EMBEDDING_DIMENSIONS_ERROR_MSG = 'An error occurred, the embedding dimensions must be positive integers and separated\n' \
                                 'by a comma e.g. 50,100,200' \
                                 'Please try again. \n'

RELATION_EMBEDDING_DIMENSION_PRINT_MSG = 'Please provide the embedding dimension of relations:'
RELATION_EMBEDDING_DIMENSION_PROMPT_MSG = '> Relation embedding dimension: '
RELATION_EMBEDDING_DIMENSION_ERROR_MSG = 'An error occurred, please type in an integer as embedding dimension such as 30.'

RELATION_EMBEDDING_DIMENSIONS_PRINT_MSG = 'Please provide (comma separated) the embedding dimensions of relations:'
RELATION_EMBEDDING_DIMENSIONS_PROMPT_MSG = '> Relation embedding dimensions: '
RELATION_EMBEDDING_DIMENSIONS_ERROR_MSG = 'An error occurred, the relation embedding dimensions must be positive integers and separated\n' \
                                          'by a comma e.g. 50,100,200' \
                                          'Please try again. \n'

MARGIN_LOSS_PRINT_MSG = 'Please type in the margin losses:'
MARGIN_LOSS_PROMPT_MSG = '> Margin loss:'
MARGIN_LOSS_ERROR_MSG = 'An error occurred, please type in a float value.'

LEARNING_RATE_PRINT_MSG = 'Please type in the learning rate.'
LEARNING_RATE_PROMPT_MSG = '> Learning rate:'
LEARNING_RATE_ERROR_MSG = 'An error occurred, the learning rate should be a positive float value.\n' \
                          'Please try again.'

BATCH_SIZE_PRINT_MSG = 'Please type the batch size comma:'
BATCH_SIZE_PROMPT_MSG = '> Batch size:'
BATCH_SIZE_ERROR_MSG = 'An error occurred, please select a integer.'

EPOCH_PRINT_MSG = 'Please type the number of epochs:'
EPOCH_PROMPT_MSG = 'Epochs'
EPOCH_ERROR_MSG = 'An error occurred, please select an integers.'

SEED_PRINT_MSG = 'Please specify the random seed.'
SEED_PROMPT_MSG = 'Random seed'
SEED_ERROR_MSG = 'An error occurred, please proive and positive integer as the random seed.'

ENTITIES_NORMALIZATION_PRINT_MSG = 'Please select a norm to use to normalize the entities. The norm should be a positive integer greater than 0'
ENTITIES_NORMALIZATION_PROMPT_MSG = '> Norm to use for normalization of the entities: '
ENTITIES_NORMALIZATION_ERROR_MSG = 'An error occurred, the norm should be an integer greater than 0, such as 1\n' \
                                   'Please try again.'

NORMS_FOR_NORMALIZATION_OF_ENTITIES_PRINT_MSG = 'Please select (comma separated) a list of norms to use for normalizing the entities.'
NORMS_FOR_NORMALIZATION_OF_ENTITIES_PROMPT_MSG = '> Norms: '
NORMS_FOR_NORMALIZATION_OF_ENTITIES_ERROR_MSG = 'An error occurred, the normalization norms should integers, greater than 0, \n' \
                                                ' and separated by a comma (e.g. 1,2,3). Please try again.'

NORM_SCORING_FUNCTION_PRINT_MSG = 'Please select a norm to use as a scoring function. The norm should be a positive integer greater than 0'
NORM_SCORING_FUNCTION_PROMPT_MSG = '> Norm to use as scoring function: '
NORM_SCORING_FUNCTION_ERROR_MSG = 'An error occurred, the norm for the scoring function should be an integer greater than 0, such as 1\n' \
                                  'Please try again.'

NORMS_SCORING_FUNCTION_PRINT_MSG = 'Please select (comma separated) a list of norms to use as a scoring functions. The norms should be a positive integers greater than 0'
NORMS_SCORING_FUNCTION_PROMPT_MSG = '> Norms to use as scoring function: '
NORMS_SCORING_FUNCTION_ERROR_MSG = 'An error occurred, the norms for the scoring functions should integers, greater than 0, \n' \
                                   ' and separated by a comma (e.g. 1,2,3). Please try again.'

SAVE_CONFIG_PRINT_MSG = 'Do you want to save the configuration file?'
SAVE_CONFIG_PROMPT_MSG = '> \'yes\' or \'no\':'
SAVE_CONFIG_ERROR_MSG = 'An error occurred, please type \'yes\' or \'no\'.'

K_FOR_HITS_AT_K_PRINT_MSG = 'Please select \'k\' for hits@k'
K_FOR_HITS_AT_K_PROMPT_MSG = '> k:'
K_FOR_HITS_AT_K_ERROR_MSG = 'An error occurred, \'k\' must be a positive integer.'

CONV_E_HPO_INPUT_CHANNELS_PRINT_MSG = 'Please select (comma seperated) the number of input channels for ConvE'
CONV_E_HPO_INPUT_CHANNELS_PROMPT_MSG = '> Input channels:'
CONV_E_HPO_INPUT_CHANNELS_ERROR_MSG = 'An error occurred, input channels must be positive integers.'

CONV_E_INPUT_CHANNEL_PRINT_MSG = 'Please select the number of input channels.'
CONV_E_INPUT_CHANNEL_PROMPT_MSG = '> Number of input channels:'
CONV_E_INPUT_CHANNEL_ERROR_MSG = 'An error occurred, the number of input channels must be a positive integer.\n' \
                                 'Please try again.'

CONV_E_HPO_OUT_CHANNELS_PRINT_MSG = 'Please select (comma seperated) the number of output channels for ConvE'
CONV_E_HPO_OUT_CHANNELS_PROMPT_MSG = '> Output channels:'
CONV_E_HPO_OUT_CHANNELS_ERROR_MSG = 'An error occurred, output channels must be positive integers.'

CONV_E_OUT_CHANNEL_PRINT_MSG = 'Please select the number of output channels.'
CONV_E_OUT_CHANNEL_PROMPT_MSG = '> Output channels:'
CONV_E_OUT_CHANNEL_ERROR_MSG = 'An error occurred, the number of output channels must be a positive number.'

CONV_E_HPO_KERNEL_HEIGHTS_PRINT_MSG = 'Please select the kernel heights for ConvE'
CONV_E_HPO_KERNEL_HEIGHTS_PROMPT_MSG = '> Kernel height for defined of height \'%d\':'
CONV_E_HPO_KERNEL_HEIGHTS_ERROR_MSG = 'An error occurred, kernel heights must be positive integers and <= than %d (defined height).'

CONV_E_KERNEL_HEIGHT_PRINT_MSG = 'Please select the height for the convolution kernel based on the specified embedding height of %d.'
CONV_E_KERNEL_HEIGHT_PROMPT_MSG = '> Convolution kernel: '
CONV_E_KERNEL_HEIGHT_ERROR_MSG = 'An error occurred, the kernel height must be a positive integer and <= than %d (defined height).'

CONV_E_HPO_KERNEL_WIDTHS_PRINT_MSG = 'Please select the wifth for the convolution kernel.'
CONV_E_HPO_KERNEL_WIDTHS_PROMPT_MSG = '> Kernel width for defined width %d: '
CONV_E_HPO_KERNEL_WIDTHS_ERROR_MSG = 'An error occurred, kernel widths mus be positive integers and <= than %d (defined width).'

CONV_E_KERNEL_WIDTH_PRINT_MSG = 'Please select the kernel width for ConvE based on the specified embedding width of %d'
CONV_E_KERNEL_WIDTH_PROMPT_MSG = '> Kernel width for defined width of \'%d\': '
CONV_E_KERNEL_WIDTH_ERROR_MSG = 'An error occurred, kernel width mus be a positive integer and <= than %d (defined width).'

TRAINING_SET_PRINT_MSG = 'Please provide the path to the training file.'
TRAINING_FILE_ERROR_MSG = 'An error occurred, either the path is not correct or the training file doesn\'t exist.\n' \
                          'Please try again.'

CONFIG_FILE_PROMPT_MSG = '> Please provide the path to your existing configuration file: '
CONFIG_FILE_ERROR_MSG = 'An error occurred, please make sure that the file exists, and that it is JSON file.\n' \
                        'Please try again.'

CONV_E_HPO_INPUT_DROPOUTS_PRINT_MSG = 'Please select (comma separated) the input dropout value(s)'
CONV_E_HPO_INPUT_DROPOUTS_PROMPT_MSG = '> Input dropout value(s):'
CONV_E_HPO_INPUT_DROPOUTS_ERROR_MSG = 'An error occurred, input must be positive float values.'

CONV_E_INPUT_DROPOUT_PRINT_MSG = 'Please select the dropout rate for the input layer.'
CONV_E_INPUT_DROPOUT_PROMPT_MSG = '> Dropout rate: '
CONV_E_INPUT_DROPOUT_ERROR_MSG = 'An error occurred, the dropout rate must be a positive float value between 0 and 1.'

CONV_E_HPO_OUTPUT_DROPOUT_PRINT_MSG = 'Please select (comma separated) the output dropout value(s)'
CONV_E_HPO_OUTPUT_DROPOUT_PROMPT_MSG = '> Output dropout value(s):'
CONV_E_HPO_OUTPUT_DROPOUT_ERROR_MSG = 'An error occurred, input must be positive float values.'

CONV_E_OUTPUT_DROPOUT_PRINT_MSG = 'Please select the output dropout value'
CONV_E_OUTPUT_DROPOUT_PROMPT_MSG = '> Output dropout value:'
CONV_E_OUTPUT_DROPOUT_ERROR_MSG = 'An error occurred, input must be positive float values.'

CONV_E_HPO_FEATURE_MAP_DROPOUT_PRINT_MSG = 'Please select (comma separated) the feature map dropout value(s)'
CONV_E_HPO_FEATURE_MAP_DROPOUT_PROMPT_MSG = '> Feature map dropout value(s):'
CONV_E_HPO_FEATURE_MAP_DROPOUT_ERROR_MSG = 'An error occurred, input must be positive float values.'

CONV_E_FEATURE_MAP_DROPOUT_PRINT_MSG = 'Please select the feature map dropout value'
CONV_E__FEATURE_MAP_DROPOUT_PROMPT_MSG = '> Feature map dropout value:'
CONV_E_FEATURE_MAP_DROPOUT_ERROR_MSG = 'Invalid output, input must be a positive float value.'

WEIGHT_SOFT_CONSTRAINT_TRANS_H_PRINT_MSG = 'Please select the weight value for the soft constraints of the loss function'
WEIGHT_SOFT_CONSTRAINT_TRANS_H_PROMPT_MSG = '> Weight value for soft constraints: '
WEIGHT_SOFT_CONSTRAINT_TRANS_H_ERROR_MSG = 'An error occurred, input must be positive a float value.'

WEIGHTS_SOFT_CONSTRAINT_TRANS_H_PRINT_MSG = 'Please provide (comma separated) the weight values for weighting the soft constraints of the loss function'
WEIGHTS_SOFT_CONSTRAINT_TRANS_H_PROMPT_MSG = '> Weight values for soft constraints: '
WEIGHTS_SOFT_CONSTRAINT_TRANS_H_ERROR_MSG = 'An error occurred, the weight values for the soft constraint must be float values, \n' \
                                            ' and separated by a comma (e.g. 0.1,0.5). Please try again.'



# -*- coding: utf-8 -*-

"""Utilities for acquiring and formatting triples."""

from .data_transformation_utils import transform_freebase_to_internal_format, transform_wn_18_to_internal_format
from .instance_creation_utils import create_mapped_triples, create_mappings



# -*- coding: utf-8 -*-

"""Implementation of DistMult."""

from typing import Dict

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import DISTMULT_NAME
from pykeen.kge_models.base import BaseModule, slice_triples

__all__ = ['DistMult']


class DistMult(BaseModule):
    """An implementation of DistMult [yang2014]_.

    This model simplifies RESCAL by restricting matrices representing relations as diagonal matrices.

    .. [yang2014] Yang, B., Yih, W., He, X., Gao, J., & Deng, L. (2014). `Embedding Entities and Relations for Learning
                  and Inference in Knowledge Bases <https://arxiv.org/pdf/1412.6575.pdf>`_. CoRR, abs/1412.6575.

    .. seealso::

       - Alternative implementation in OpenKE: https://github.com/thunlp/OpenKE/blob/master/models/DistMult.py
    """

    model_name = DISTMULT_NAME
    margin_ranking_loss_size_average: bool = True

    def __init__(self, config: Dict) -> None:
        super().__init__(config)

        # Embeddings
        self.relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim)

        self._initialize()

    def _initialize(self):
        # The same bound is used for both entity embeddings and relation embeddings because they have the same dimension
        embeddings_init_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(
            self.entity_embeddings.weight.data,
            a=-embeddings_init_bound,
            b=+embeddings_init_bound,
        )
        nn.init.uniform_(
            self.relation_embeddings.weight.data,
            a=-embeddings_init_bound,
            b=+embeddings_init_bound,
        )

        norms = torch.norm(self.relation_embeddings.weight, p=2, dim=1).data
        self.relation_embeddings.weight.data = self.relation_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.relation_embeddings.weight))

    def predict(self, triples):
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)
        scores = self._score_triples(triples)
        return scores.detach().cpu().numpy()

    def forward(self, positives, negatives):
        positive_scores = self._score_triples(positives)
        negative_scores = self._score_triples(negatives)
        loss = self._compute_loss(positive_scores=positive_scores, negative_scores=negative_scores)
        return loss

    def _compute_loss(self, positive_scores, negative_scores):
        # Choose y = -1 since a smaller score is better.
        # In TransE for example, the scores represent distances
        y = np.repeat([-1], repeats=positive_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        positive_scores = torch.tensor(positive_scores, dtype=torch.float, device=self.device)
        negative_scores = torch.tensor(negative_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(positive_scores, negative_scores, y)
        return loss

    def _score_triples(self, triples):
        head_embeddings, relation_embeddings, tail_embeddings = self._get_triple_embeddings(triples)
        return self._compute_scores(head_embeddings, relation_embeddings, tail_embeddings)

    @staticmethod
    def _compute_scores(head_embeddings, relation_embeddings, tail_embeddings):
        scores = - torch.sum(head_embeddings * relation_embeddings * tail_embeddings, dim=1)
        return scores

    def _get_triple_embeddings(self, triples):
        heads, relations, tails = slice_triples(triples)
        return (
            self._get_entity_embeddings(heads),
            self._get_relation_embeddings(relations),
            self._get_entity_embeddings(tails)
        )

    def _get_relation_embeddings(self, relations):
        return self.relation_embeddings(relations).view(-1, self.embedding_dim)



# -*- coding: utf-8 -*-

"""Implementation of the TransE model."""

import logging
from dataclasses import dataclass
from typing import Dict

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import NORM_FOR_NORMALIZATION_OF_ENTITIES, SCORING_FUNCTION_NORM, TRANS_E_NAME
from pykeen.kge_models.base import BaseModule, slice_triples

__all__ = [
    'TransE',
    'TransEConfig',
]

log = logging.getLogger(__name__)


@dataclass
class TransEConfig:
    lp_norm: str
    scoring_function_norm: str

    @classmethod
    def from_dict(cls, config: Dict) -> 'TransEConfig':
        """Generate an instance from a dictionary."""
        return cls(
            lp_norm=config[NORM_FOR_NORMALIZATION_OF_ENTITIES],
            scoring_function_norm=config[SCORING_FUNCTION_NORM],
        )


class TransE(BaseModule):
    """An implementation of TransE [borders2013]_.

     This model considers a relation as a translation from the head to the tail entity.

    .. [borders2013] Bordes, A., *et al.* (2013). `Translating embeddings for modeling multi-relational data
                     <http://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data.pdf>`_
                     . NIPS.

    .. seealso::

       - Alternative implementation in OpenKE: https://github.com/thunlp/OpenKE/blob/OpenKE-PyTorch/models/TransE.py
    """

    model_name = TRANS_E_NAME
    margin_ranking_loss_size_average: bool = True
    hyper_params = BaseModule.hyper_params + [SCORING_FUNCTION_NORM, NORM_FOR_NORMALIZATION_OF_ENTITIES]

    def __init__(self, config: Dict) -> None:
        super().__init__(config)
        config = TransEConfig.from_dict(config)

        # Embeddings
        self.l_p_norm_entities = config.lp_norm
        self.scoring_fct_norm = config.scoring_function_norm
        self.relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim)

        self._initialize()

    def _initialize(self):
        embeddings_init_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(
            self.entity_embeddings.weight.data,
            a=-embeddings_init_bound,
            b=+embeddings_init_bound,
        )
        nn.init.uniform_(
            self.relation_embeddings.weight.data,
            a=-embeddings_init_bound,
            b=+embeddings_init_bound,
        )

        norms = torch.norm(self.relation_embeddings.weight, p=2, dim=1).data
        self.relation_embeddings.weight.data = self.relation_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.relation_embeddings.weight))

    def predict(self, triples):
        scores = self._score_triples(triples)
        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        # Normalize embeddings of entities
        norms = torch.norm(self.entity_embeddings.weight, p=self.l_p_norm_entities, dim=1).data
        self.entity_embeddings.weight.data = self.entity_embeddings.weight.data.div(
            norms.view(self.num_entities, 1).expand_as(self.entity_embeddings.weight))

        positive_scores = self._score_triples(batch_positives)
        negative_scores = self._score_triples(batch_negatives)
        loss = self._compute_loss(positive_scores=positive_scores, negative_scores=negative_scores)
        return loss

    def _compute_loss(self, positive_scores, negative_scores):
        y = np.repeat([-1], repeats=positive_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the positive and negative triples
        positive_scores = torch.tensor(positive_scores, dtype=torch.float, device=self.device)
        negative_scores = torch.tensor(negative_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(positive_scores, negative_scores, y)
        return loss

    def _score_triples(self, triples):
        head_embeddings, relation_embeddings, tail_embeddings = self._get_triple_embeddings(triples)
        scores = self._compute_scores(head_embeddings, relation_embeddings, tail_embeddings)
        return scores

    def _compute_scores(self, head_embeddings, relation_embeddings, tail_embeddings):
        """Compute the scores based on the head, relation, and tail embeddings.

        :param head_embeddings: embeddings of head entities of dimension batchsize x embedding_dim
        :param relation_embeddings: emebddings of relation embeddings of dimension batchsize x embedding_dim
        :param tail_embeddings: embeddings of tail entities of dimension batchsize x embedding_dim
        :return: Tensor of dimension batch_size containing the scores for each batch element
        """
        # Add the vector element wise
        sum_res = head_embeddings + relation_embeddings - tail_embeddings
        distances = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        return distances

    def _get_triple_embeddings(self, triples):
        heads, relations, tails = slice_triples(triples)
        return (
            self._get_entity_embeddings(heads),
            self._get_relation_embeddings(relations),
            self._get_entity_embeddings(tails),
        )

    def _get_relation_embeddings(self, relations):
        return self.relation_embeddings(relations).view(-1, self.embedding_dim)



# -*- coding: utf-8 -*-

"""Implementation of structured model (SE)."""

import logging
from typing import Dict

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import NORM_FOR_NORMALIZATION_OF_ENTITIES, SCORING_FUNCTION_NORM, SE_NAME
from .base import BaseModule, slice_triples
from .trans_e import TransEConfig

__all__ = [
    'StructuredEmbedding',
]

log = logging.getLogger(__name__)


class StructuredEmbedding(BaseModule):
    """An implementation of Structured Embedding (SE) [bordes2011]_.

    This model projects different matrices for each relation head and tail entity.

    .. [bordes2011] Bordes, A., *et al.* (2011). `Learning Structured Embeddings of Knowledge Bases
                    <http://www.aaai.org/ocs/index.php/AAAI/AAAI11/paper/download/3659/3898>`_. AAAI. Vol. 6. No. 1.
    """

    model_name = SE_NAME
    margin_ranking_loss_size_average: bool = True
    hyper_params = BaseModule.hyper_params + [SCORING_FUNCTION_NORM, NORM_FOR_NORMALIZATION_OF_ENTITIES]

    def __init__(self, config: Dict) -> None:
        super().__init__(config)
        config = TransEConfig.from_dict(config)

        # Embeddings
        self.l_p_norm_entities = config.lp_norm
        self.scoring_fct_norm = config.scoring_function_norm

        self.left_relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim * self.embedding_dim)
        self.right_relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim * self.embedding_dim)

        self._initialize()

    def _initialize(self):
        entity_embeddings_init_bound = left_relation_embeddings_init_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(
            self.entity_embeddings.weight.data,
            a=-entity_embeddings_init_bound,
            b=+entity_embeddings_init_bound,
        )
        nn.init.uniform_(
            self.left_relation_embeddings.weight.data,
            a=-left_relation_embeddings_init_bound,
            b=+left_relation_embeddings_init_bound,
        )

        # FIXME @mehdi why aren't the right relation embeddings initialized?

        norms = torch.norm(self.left_relation_embeddings.weight, p=2, dim=1).data
        self.left_relation_embeddings.weight.data = self.left_relation_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.left_relation_embeddings.weight))

    def predict(self, triples):
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)
        scores = self._score_triples(triples)
        return scores.detach().cpu().numpy()

    def forward(self, positives, negatives):
        # Normalise embeddings of entities
        norms = torch.norm(self.entity_embeddings.weight, p=self.l_p_norm_entities, dim=1).data
        self.entity_embeddings.weight.data = self.entity_embeddings.weight.data.div(
            norms.view(self.num_entities, 1).expand_as(self.entity_embeddings.weight))

        positive_scores = self._score_triples(positives)
        negative_scores = self._score_triples(negatives)

        loss = self._compute_loss(positive_scores, negative_scores)
        return loss

    def _compute_loss(self, positive_scores, negative_scores):
        y = np.repeat([-1], repeats=positive_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the positive and negative triples
        positive_scores = torch.tensor(positive_scores, dtype=torch.float, device=self.device)
        negative_scores = torch.tensor(negative_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(positive_scores, negative_scores, y)
        return loss

    def _score_triples(self, triples):
        heads, relations, tails = slice_triples(triples)

        head_embeddings = self._get_entity_embeddings(heads)
        left_relation_embeddings = self._get_left_relation_embeddings(relations)
        projected_head_embeddings = self._project_entities(
            entity_embeddings=head_embeddings,
            relation_embeddings=left_relation_embeddings,
        )

        tail_embeddings = self._get_entity_embeddings(tails)
        right_relation_embeddings = self._get_right_relation_embeddings(relations)
        projected_tails_embeddings = self._project_entities(
            entity_embeddings=tail_embeddings,
            relation_embeddings=right_relation_embeddings,
        )

        scores = self._compute_scores(projected_head_embeddings, projected_tails_embeddings)
        return scores

    def _compute_scores(self, projected_head_embeddings, projected_tail_embeddings):
        difference = projected_head_embeddings - projected_tail_embeddings
        distances = torch.norm(difference, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        return distances

    def _project_entities(self, entity_embeddings, relation_embeddings):
        entity_embeddings = entity_embeddings.unsqueeze(-1)
        projected_entity_embs = torch.matmul(relation_embeddings, entity_embeddings)
        return projected_entity_embs

    def _get_left_relation_embeddings(self, relations):
        return self.left_relation_embeddings(relations).view(-1, self.embedding_dim, self.embedding_dim)

    def _get_right_relation_embeddings(self, relations):
        return self.right_relation_embeddings(relations).view(-1, self.embedding_dim, self.embedding_dim)



# -*- coding: utf-8 -*-

"""Test pipeline module."""

import json
import os
import unittest

from pykeen.constants import (
    EXECUTION_MODE,
    HPO_MODE,
    PREFERRED_DEVICE,
    SEED,
    TEST_SET_PATH,
    TEST_SET_RATIO,
    TRAINING_MODE,
    CPU)
from pykeen.utilities.pipeline import Pipeline

CONFIG = {
    SEED: 2,
    PREFERRED_DEVICE: CPU
}


class TestPipeline(unittest.TestCase):

    def setUp(self):
        self.config = CONFIG
        self.p = Pipeline(config=self.config)

    def test_instantiate_pipeline(self):
        """Test that Pipeline can be instantiated."""
        p = Pipeline(config=CONFIG)
        self.assertIsNotNone(p)

    def test_use_hpo(self):
        """Test whether execution mode is identified correctly."""
        self.p.config[EXECUTION_MODE] = HPO_MODE
        value = self.p._use_hpo(config=self.p.config)
        self.assertTrue(value)

        self.p.config[EXECUTION_MODE] = TRAINING_MODE
        value = self.p._use_hpo(config=self.p.config)
        self.assertFalse(value)

    def test_is_evaluation_required(self):
        """Test whether evaluation option is identified correctly."""
        self.p.config[TEST_SET_PATH] = '/test/path'
        value = self.p.is_evaluation_required
        self.assertTrue(value)

        del self.p.config[TEST_SET_PATH]

        self.p.config[TEST_SET_RATIO] = 0.1
        value = self.p.is_evaluation_required
        self.assertTrue(value)

        del self.p.config[TEST_SET_RATIO]

        value = self.p.is_evaluation_required
        self.assertFalse(value)

    def test_run(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))

        config_path = os.path.join(
            os.path.abspath(os.path.join(dir_path, os.pardir)),
            'test_resources',
            'configuration_training_without_eval.json'
        )

        with open(config_path) as json_data:
            config = json.load(json_data)

        config['training_set_path'] = os.path.join(
            os.path.abspath(os.path.join(dir_path, os.pardir)),
            'test_resources',
            'example_training.tsv'
        )

        config['test_set_path'] = os.path.join(
            os.path.abspath(os.path.join(dir_path, os.pardir)),
            'test_resources',
            'example_test.tsv'
        )

        self.p.config = config
        results = self.p.run()
        self.assertIsNotNone(results)


        config_path = os.path.join(
            os.path.abspath(os.path.join(dir_path, os.pardir)),
            'test_resources',
            'configuration_hpo.json'
        )

        with open(config_path) as json_data:
            config = json.load(json_data)

        config['training_set_path'] = os.path.join(
            os.path.abspath(os.path.join(dir_path, os.pardir)),
            'test_resources',
            'example_training.tsv'
        )

        config['test_set_path'] = os.path.join(
            os.path.abspath(os.path.join(dir_path, os.pardir)),
            'test_resources',
            'example_test.tsv'
        )

        self.p.config = config
        results = self.p.run()
        self.assertIsNotNone(results)


# -*- coding: utf-8 -*-

"""Implementation of ERMLP."""

from typing import Dict

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import ERMLP_NAME
from pykeen.kge_models.base import BaseModule, slice_triples

__all__ = ['ERMLP']


class ERMLP(BaseModule):
    """An implementation of ERMLP [dong2014]_.

    This model uses a neural network-based approach.

    .. [dong2014] Dong, X., *et al.* (2014) `Knowledge vault: A web-scale approach to probabilistic knowledge fusion
                  <https://dl.acm.org/citation.cfm?id=2623623>`_. ACM.
    """

    model_name = ERMLP_NAME
    margin_ranking_loss_size_average: bool = False

    def __init__(self, config: Dict) -> None:
        super().__init__(config)

        #: Embeddings for relations in the knowledge graph
        self.relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim)

        # TODO document what is MLP?
        #:
        self.mlp = nn.Sequential(
            nn.Linear(3 * self.embedding_dim, self.embedding_dim),
            nn.ReLU(),
            nn.Dropout(p=0.2),
            nn.Linear(self.embedding_dim, 1),
        )

    def predict(self, triples):
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)
        scores = self._score_triples(triples)
        return scores.detach().cpu().numpy()

    def forward(self, positives, negatives):
        positive_scores = self._score_triples(positives)
        negative_scores = self._score_triples(negatives)
        loss = self._compute_loss(positive_scores, negative_scores)
        return loss

    def _compute_loss(self, positive_scores, negative_scores):
        y = np.repeat([-1], repeats=positive_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        positive_scores = torch.tensor(positive_scores, dtype=torch.float, device=self.device)
        negative_scores = torch.tensor(negative_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(positive_scores, negative_scores, y)
        return loss

    def _score_triples(self, triples):
        head_embeddings, relation_embeddings, tail_embeddings = self._get_triple_embeddings(triples)
        scores = self._compute_scores(head_embeddings, relation_embeddings, tail_embeddings)
        return scores

    def _compute_scores(self, head_embeddings, relation_embeddings, tail_embeddings):
        x_s = torch.cat([head_embeddings, relation_embeddings, tail_embeddings], 1)
        scores = - self.mlp(x_s)
        return scores

    def _get_triple_embeddings(self, triples):
        heads, relations, tails = slice_triples(triples)
        return (
            self._get_entity_embeddings(heads),
            self._get_relation_embeddings(relations),
            self._get_entity_embeddings(tails),
        )

    def _get_relation_embeddings(self, relations):
        return self.relation_embeddings(relations).view(-1, self.embedding_dim)



# -*- coding: utf-8 -*-

"""Constants defined for PyKEEN's CLI."""

from collections import OrderedDict

from pykeen.constants import ADAGRAD_OPTIMIZER_NAME, ADAM_OPTIMIZER_NAME, SGD_OPTIMIZER_NAME
from ...kge_models import (
    ConvE, DistMult, ERMLP, RESCAL, StructuredEmbedding, TransD, TransE, TransH, TransR, UnstructuredModel,
)

__all__ = [
    'ID_TO_KG_MODEL_MAPPING',
    'KG_MODEL_TO_ID_MAPPING',
    'ID_TO_OPTIMIZER_MAPPING',
    'OPTIMIZER_TO_ID_MAPPING',
]

ID_TO_KG_MODEL_MAPPING = OrderedDict({
    '1': TransE.model_name,
    '2': TransH.model_name,
    '3': TransR.model_name,
    '4': TransD.model_name,
    '5': StructuredEmbedding.model_name,
    '6': UnstructuredModel.model_name,
    '7': DistMult.model_name,
    '8': ERMLP.model_name,
    '9': RESCAL.model_name,
    '10': ConvE.model_name,
})

KG_MODEL_TO_ID_MAPPING = OrderedDict({
    value: key
    for key, value in ID_TO_KG_MODEL_MAPPING.items()
})

ID_TO_OPTIMIZER_MAPPING = OrderedDict({
    '1': SGD_OPTIMIZER_NAME,
    '2': ADAGRAD_OPTIMIZER_NAME,
    '3': ADAM_OPTIMIZER_NAME,
})

OPTIMIZER_TO_ID_MAPPING = OrderedDict({
    value: key
    for key, value in ID_TO_OPTIMIZER_MAPPING.items()
})



# -*- coding: utf-8 -*-

"""Test KGE models"""

import unittest

from pykeen.constants import *
from pykeen.kge_models import TransE, TransH, TransR, TransD, DistMult, ERMLP, StructuredEmbedding, UnstructuredModel, \
    RESCAL, ConvE

import torch

TRANS_E_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    NORM_FOR_NORMALIZATION_OF_ENTITIES: 2,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
}

TRANS_H_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    NORM_FOR_NORMALIZATION_OF_ENTITIES: 2,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
    WEIGHT_SOFT_CONSTRAINT_TRANS_H: 0.05,
}

TRANS_R_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    RELATION_EMBEDDING_DIM: 3,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
}

TRANS_D_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    RELATION_EMBEDDING_DIM: 3,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
}

DISTMULT_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
}

ERMLP_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 2,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
}

SE_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    NORM_FOR_NORMALIZATION_OF_ENTITIES: 2,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
}

UM_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    NORM_FOR_NORMALIZATION_OF_ENTITIES: 2,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
}

RESCAL_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    NORM_FOR_NORMALIZATION_OF_ENTITIES: 2,
    SCORING_FUNCTION_NORM: 1,
    MARGIN_LOSS: 4,
}

CONV_E_CONFIG = {
    NUM_ENTITIES: 5,
    NUM_RELATIONS: 5,
    EMBEDDING_DIM: 5,
    CONV_E_INPUT_CHANNELS: 1,
    CONV_E_OUTPUT_CHANNELS: 2,
    CONV_E_KERNEL_HEIGHT: 5,
    CONV_E_KERNEL_WIDTH: 1,
    CONV_E_INPUT_DROPOUT: 0.3,
    CONV_E_OUTPUT_DROPOUT: 0.5,
    CONV_E_FEATURE_MAP_DROPOUT: 0.2,
    CONV_E_HEIGHT: 5,
    CONV_E_WIDTH: 1,
}


class TestModelInstantiation(unittest.TestCase):
    """Test that all models can be instantiated."""

    def test_instantiate_trans_e(self):
        """Test that TransE can be instantiated."""
        trans_e = TransE(config=TRANS_E_CONFIG)
        self.assertIsNotNone(trans_e)
        self.assertEqual(trans_e.num_entities, 5)
        self.assertEqual(trans_e.num_relations, 5)
        self.assertEqual(trans_e.embedding_dim, 5)
        self.assertEqual(trans_e.l_p_norm_entities, 2)
        self.assertEqual(trans_e.scoring_fct_norm, 1)
        self.assertEqual(trans_e.margin_loss, 4)

    def test_instantiate_trans_h(self):
        """Test that TransH can be instantiated."""
        trans_h = TransH(config=TRANS_H_CONFIG)
        self.assertIsNotNone(trans_h)
        self.assertEqual(trans_h.num_entities, 5)
        self.assertEqual(trans_h.num_relations, 5)
        self.assertEqual(trans_h.embedding_dim, 5)
        self.assertEqual(trans_h.weighting_soft_constraint, 0.05)
        self.assertEqual(trans_h.scoring_fct_norm, 1)
        self.assertEqual(trans_h.margin_loss, 4)

    def test_instantiate_trans_r(self):
        """Test that TransR can be instantiated."""
        trans_r = TransR(config=TRANS_R_CONFIG)
        self.assertIsNotNone(trans_r)
        self.assertEqual(trans_r.num_entities, 5)
        self.assertEqual(trans_r.num_relations, 5)
        self.assertEqual(trans_r.embedding_dim, 5)
        self.assertEqual(trans_r.relation_embedding_dim, 3)
        self.assertEqual(trans_r.scoring_fct_norm, 1)
        self.assertEqual(trans_r.margin_loss, 4)

    def test_instantiate_trans_d(self):
        """Test that TransD can be instantiated."""
        trans_d = TransD(config=TRANS_D_CONFIG)
        self.assertIsNotNone(trans_d)
        self.assertEqual(trans_d.num_entities, 5)
        self.assertEqual(trans_d.num_relations, 5)
        self.assertEqual(trans_d.embedding_dim, 5)
        self.assertEqual(trans_d.relation_embedding_dim, 3)
        self.assertEqual(trans_d.scoring_fct_norm, 1)
        self.assertEqual(trans_d.margin_loss, 4)

    def test_instantiate_distmult(self):
        """Test that DistMult can be instantiated."""
        distmult = DistMult(config=DISTMULT_CONFIG)
        self.assertIsNotNone(distmult)
        self.assertEqual(distmult.num_entities, 5)
        self.assertEqual(distmult.num_relations, 5)
        self.assertEqual(distmult.embedding_dim, 5)
        self.assertEqual(distmult.margin_loss, 4)

    def test_instantiate_ermlp(self):
        """Test that ERMLP can be instantiated."""
        ermlp = ERMLP(config=ERMLP_CONFIG)
        self.assertIsNotNone(ermlp)
        self.assertEqual(ermlp.num_entities, 5)
        self.assertEqual(ermlp.num_relations, 5)
        self.assertEqual(ermlp.embedding_dim, 2)
        self.assertEqual(ermlp.margin_loss, 4)

    def test_instantiate_strcutured_embedding(self):
        """Test that StructuredEmbedding can be instantiated."""
        se = StructuredEmbedding(config=SE_CONFIG)
        self.assertIsNotNone(se)
        self.assertEqual(se.num_entities, 5)
        self.assertEqual(se.num_relations, 5)
        self.assertEqual(se.embedding_dim, 5)
        self.assertEqual(se.l_p_norm_entities, 2)
        self.assertEqual(se.margin_loss, 4)

    def test_instantiate_unstructured_model(self):
        """Test that UnstructuredModel can be instantiated."""
        um = UnstructuredModel(config=UM_CONFIG)
        self.assertIsNotNone(um)
        self.assertEqual(um.num_entities, 5)
        self.assertEqual(um.num_relations, 5)
        self.assertEqual(um.embedding_dim, 5)
        self.assertEqual(um.l_p_norm_entities, 2)
        self.assertEqual(um.margin_loss, 4)

    def test_instantiate_rescal(self):
        """Test that RESCAL can be instantiated."""
        rescal = RESCAL(config=RESCAL_CONFIG)
        self.assertIsNotNone(rescal)
        self.assertEqual(rescal.num_entities, 5)
        self.assertEqual(rescal.num_relations, 5)
        self.assertEqual(rescal.embedding_dim, 5)
        self.assertEqual(rescal.margin_loss, 4)

    def test_instantiate_conv_e(self):
        """Test that ConvE can be instantiated."""
        conv_e = ConvE(config=CONV_E_CONFIG)
        self.assertIsNotNone(conv_e)
        self.assertEqual(conv_e.num_entities, 5)
        self.assertEqual(conv_e.num_relations, 5)
        self.assertEqual(conv_e.embedding_dim, 5)


class TestScoringFunctions(unittest.TestCase):
    def test_compute_scores_trans_e(self):
        """Test that TransE's socore function computes the scores correct."""
        trans_e = TransE(config=TRANS_E_CONFIG)
        h_embs = torch.tensor([[1.,1.],[2.,2.]],dtype=torch.float)
        r_embs = torch.tensor([[1., 1.], [2., 2.]],dtype=torch.float)
        t_embs = torch.tensor([[2., 2.], [4., 4.]],dtype=torch.float)

        scores = trans_e._compute_scores(h_embs,r_embs,t_embs).cpu().numpy().tolist()

        self.assertEqual(scores, [0.,0.])

    def test_compute_scores_trans_h(self):
        """Test that TransH's socore function computes the scores correct."""
        trans_h = TransH(config=TRANS_H_CONFIG)
        proj_h_embs = torch.tensor([[1.,1.],[1.,1.]],dtype=torch.float)
        proj_r_embs = torch.tensor([[1., 1.], [2., 2.]],dtype=torch.float)
        proj_t_embs = torch.tensor([[2., 2.], [4., 4.]],dtype=torch.float)

        scores = trans_h._compute_scores(proj_h_embs,proj_r_embs,proj_t_embs).cpu().numpy().tolist()

        self.assertEqual(scores, [0.,4.])

    def test_compute_scores_trans_r(self):
        """Test that TransR's socore function computes the scores correct."""
        trans_r = TransR(config=TRANS_R_CONFIG)
        proj_h_embs = torch.tensor([[1.,1.],[1.,1.]],dtype=torch.float)
        proj_r_embs = torch.tensor([[1., 1.], [2., 2.]],dtype=torch.float)
        proj_t_embs = torch.tensor([[2., 2.], [4., 4.]],dtype=torch.float)

        scores = trans_r._compute_scores(proj_h_embs,proj_r_embs,proj_t_embs).cpu().numpy().tolist()

        self.assertEqual(scores, [0.,4.])

    def test_compute_scores_trans_d(self):
        """Test that TransD's socore function computes the scores correct."""
        trans_d = TransD(config=TRANS_D_CONFIG)
        proj_h_embs = torch.tensor([[1.,1.],[1.,1.]],dtype=torch.float)
        proj_r_embs = torch.tensor([[1., 1.], [2., 2.]],dtype=torch.float)
        proj_t_embs = torch.tensor([[2., 2.], [4., 4.]],dtype=torch.float)

        scores = trans_d._compute_scores(proj_h_embs,proj_r_embs,proj_t_embs).cpu().numpy().tolist()

        self.assertEqual(scores, [0.,4.])

    def test_compute_scores_distmult(self):
        """Test that DistMult's socore function computes the scores correct."""
        distmult = DistMult(config=DISTMULT_CONFIG)
        h_embs = torch.tensor([[1.,1.],[1.,1.]],dtype=torch.float)
        r_embs = torch.tensor([[1., 1.], [2., 2.]],dtype=torch.float)
        t_embs = torch.tensor([[2., 2.], [4., 4.]],dtype=torch.float)

        scores = distmult._compute_scores(h_embs,r_embs,t_embs).cpu().numpy().tolist()

        self.assertEqual(scores, [-4.,-16.])

    def test_compute_scores_um(self):
        """Test that DistMult's socore function computes the scores correct."""
        um = UnstructuredModel(config=UM_CONFIG)
        h_embs = torch.tensor([[1.,1.],[1.,1.]],dtype=torch.float)
        t_embs = torch.tensor([[2., 2.], [4., 4.]],dtype=torch.float)

        scores = um._compute_scores(h_embs, t_embs).cpu().numpy().tolist()

        self.assertEqual(scores, [4.,36.])

    def test_compute_scores_se(self):
        """Test that SE's socore function computes the scores correct."""
        se = StructuredEmbedding(config=SE_CONFIG)
        proj_h_embs = torch.tensor([[1.,1.],[1.,1.]],dtype=torch.float)
        proj_t_embs = torch.tensor([[2., 2.], [4., 4.]],dtype=torch.float)

        scores = se._compute_scores(proj_h_embs,proj_t_embs).cpu().numpy().tolist()

        self.assertEqual(scores, [2.,6.])

    def test_compute_scores_ermlp(self):
        """Test that SE's socore function computes the scores correct."""
        ermlp = ERMLP(config=ERMLP_CONFIG)

        h_embs = torch.tensor([[1., 1.], [1., 1.]], dtype=torch.float)
        r_embs = torch.tensor([[1., 1.], [2., 2.]], dtype=torch.float)
        t_embs = torch.tensor([[2., 2.], [4., 4.]], dtype=torch.float)

        scores = ermlp._compute_scores(h_embs,r_embs,t_embs).detach().cpu().numpy().tolist()

        self.assertEqual(len(scores),2)



